@@ -24,6 +24,10 @@
  */
 #include "slirp.h"
 
+#ifndef g_warning_once
+#define g_warning_once g_warning
+#endif
+
 
 #ifndef _WIN32
 #include <net/if.h>
@@ -65,7 +69,7 @@ static const struct in6_addr SITE_LOCAL_DNS_BROADCAST_ADDRS[] = {
 
 #endif
 
-int slirp_debug;
+unsigned int slirp_debug;
 
 /* Define to 1 if you want KEEPALIVE timers */
 bool slirp_do_keepalive;
@@ -139,7 +143,7 @@ int get_dns_addr(struct in_addr *pdns_addr)
     return 0;
 }
 
-int is_site_local_dns_broadcast(struct in6_addr *address)
+static int is_site_local_dns_broadcast(struct in6_addr *address)
 {
     int i;
     for (i = 0; i < G_N_ELEMENTS(SITE_LOCAL_DNS_BROADCAST_ADDRS); i++) {
@@ -150,7 +154,7 @@ int is_site_local_dns_broadcast(struct in6_addr *address)
     return 0;
 }
 
-void print_dns_v6_address(struct in6_addr address)
+static void print_dns_v6_address(struct in6_addr address)
 {
     char address_str[INET6_ADDRSTRLEN] = "";
     if (inet_ntop(AF_INET6, &address, address_str, INET6_ADDRSTRLEN)
@@ -158,7 +162,7 @@ void print_dns_v6_address(struct in6_addr address)
         DEBUG_ERROR("Failed to stringify IPv6 address for logging.");
         return;
     }
-    DEBUG_CALL("IPv6 DNS server found: %s", address_str);
+    DEBUG_RAW_CALL("IPv6 DNS server found: %s", address_str);
 }
 
 // Gets the first valid DNS resolver with an IPv6 address.
@@ -166,7 +170,7 @@ void print_dns_v6_address(struct in6_addr address)
 // are on deprecated addresses and not generally expected
 // to work. Further details at:
 // https://www.ietf.org/proceedings/52/I-D/draft-ietf-ipngwg-dns-discovery-03.txt
-int get_ipv6_dns_server(struct in6_addr *dns_server_address, uint32_t *scope_id)
+static int get_ipv6_dns_server(struct in6_addr *dns_server_address, uint32_t *scope_id)
 {
     PIP_ADAPTER_ADDRESSES addresses = NULL;
     PIP_ADAPTER_ADDRESSES address = NULL;
@@ -285,9 +289,12 @@ static int get_dns_addr_libresolv(int af, void *pdns_addr, void *cached_addr,
     found = 0;
     DEBUG_MISC("IP address of your DNS(s):");
     for (int i = 0; i < count; i++) {
-        if (af == servers[i].sin.sin_family) {
-            found++;
+        if (af != servers[i].sin.sin_family) {
+            continue;
         }
+
+        found++;
+
         if (af == AF_INET) {
             addr = &servers[i].sin.sin_addr;
         } else { // af == AF_INET6
@@ -310,7 +317,7 @@ static int get_dns_addr_libresolv(int af, void *pdns_addr, void *cached_addr,
         if (found > 3) {
             DEBUG_MISC("  (more)");
             break;
-        } else if (slirp_debug & DBG_MISC) {
+        } else if (slirp_debug & SLIRP_DBG_MISC) {
             char s[INET6_ADDRSTRLEN];
             const char *res = inet_ntop(af, addr, s, sizeof(s));
             if (!res) {
@@ -389,6 +396,48 @@ static int get_dns_addr_cached(void *pdns_addr, void *cached_addr,
     return 1;
 }
 
+static bool try_and_setdns_server(int af, unsigned found, unsigned if_index,
+	const char *buff2, void *pdns_addr, void *cached_addr,
+	socklen_t addrlen, uint32_t *scope_id, uint32_t *cached_scope_id,
+	unsigned *cached_time)
+{
+    union {
+        struct in_addr dns_addr;
+        struct in6_addr dns6_addr;
+    } tmp_addr;
+
+    assert(sizeof(tmp_addr) >= addrlen);
+
+    if (!inet_pton(af, buff2, &tmp_addr))
+	return false;
+
+    /* If it's the first one, set it to dns_addr */
+    if (!found) {
+	memcpy(pdns_addr, &tmp_addr, addrlen);
+	memcpy(cached_addr, &tmp_addr, addrlen);
+	if (scope_id) {
+	    *scope_id = if_index;
+	}
+	if (cached_scope_id) {
+	    *cached_scope_id = if_index;
+	}
+	*cached_time = curtime;
+    }
+
+    if (found > 2) {
+	DEBUG_MISC("  (more)");
+    } else if (slirp_debug & SLIRP_DBG_MISC) {
+	char s[INET6_ADDRSTRLEN];
+	const char *res = inet_ntop(af, &tmp_addr, s, sizeof(s));
+	if (!res) {
+	    res = "  (string conversion error)";
+	}
+	DEBUG_MISC("  %s", res);
+    }
+
+    return true;
+}
+
 static int get_dns_addr_resolv_conf(int af, void *pdns_addr, void *cached_addr,
                                     socklen_t addrlen,
                                     uint32_t *scope_id, uint32_t *cached_scope_id,
@@ -398,13 +447,9 @@ static int get_dns_addr_resolv_conf(int af, void *pdns_addr, void *cached_addr,
     char buff2[257];
     FILE *f;
     int found = 0;
-    union {
-        struct in_addr dns_addr;
-        struct in6_addr dns6_addr;
-    } tmp_addr;
     unsigned if_index;
+    unsigned nameservers = 0;
 
-    assert(sizeof(tmp_addr) >= addrlen);
     f = fopen(RESOLV_CONF_PATH, "r");
     if (!f)
         return -1;
@@ -420,39 +465,30 @@ static int get_dns_addr_resolv_conf(int af, void *pdns_addr, void *cached_addr,
                 if_index = 0;
             }
 
-            if (!inet_pton(af, buff2, &tmp_addr)) {
-                continue;
-            }
-            /* If it's the first one, set it to dns_addr */
-            if (!found) {
-                memcpy(pdns_addr, &tmp_addr, addrlen);
-                memcpy(cached_addr, &tmp_addr, addrlen);
-                if (scope_id) {
-                    *scope_id = if_index;
-                }
-                if (cached_scope_id) {
-                    *cached_scope_id = if_index;
-                }
-                *cached_time = curtime;
-            }
+	    nameservers++;
 
-            if (++found > 3) {
-                DEBUG_MISC("  (more)");
-                break;
-            } else if (slirp_debug & DBG_MISC) {
-                char s[INET6_ADDRSTRLEN];
-                const char *res = inet_ntop(af, &tmp_addr, s, sizeof(s));
-                if (!res) {
-                    res = "  (string conversion error)";
-                }
-                DEBUG_MISC("  %s", res);
-            }
+	    if (!try_and_setdns_server(af, found, if_index, buff2, pdns_addr,
+				    cached_addr, addrlen, scope_id,
+				    cached_scope_id, cached_time))
+		    continue;
+
+	    if (++found > 3)
+		break;
         }
     }
     fclose(f);
-    if (!found)
+    if (nameservers && !found)
         return -1;
-    return 0;
+    if (!nameservers) {
+	found += try_and_setdns_server(af, found, 0, "127.0.0.1",
+			pdns_addr, cached_addr, addrlen, scope_id,
+			cached_scope_id, cached_time);
+	found += try_and_setdns_server(af, found, 0, "::1",
+			pdns_addr, cached_addr, addrlen, scope_id,
+			cached_scope_id, cached_time);
+    }
+
+    return found ? 0 : -1;
 }
 
 int get_dns_addr(struct in_addr *pdns_addr)
@@ -518,11 +554,11 @@ static void slirp_init_once(void)
     debug = g_getenv("SLIRP_DEBUG");
     if (debug) {
         const GDebugKey keys[] = {
-            { "call", DBG_CALL },
-            { "misc", DBG_MISC },
-            { "error", DBG_ERROR },
-            { "tftp", DBG_TFTP },
-            { "verbose_call", DBG_VERBOSE_CALL },
+            { "call", SLIRP_DBG_CALL },
+            { "misc", SLIRP_DBG_MISC },
+            { "error", SLIRP_DBG_ERROR },
+            { "tftp", SLIRP_DBG_TFTP },
+            { "verbose_call", SLIRP_DBG_VERBOSE_CALL },
         };
         slirp_debug = g_parse_debug_string(debug, keys, G_N_ELEMENTS(keys));
     }
@@ -532,7 +568,7 @@ static void ra_timer_handler_cb(void *opaque)
 {
     Slirp *slirp = opaque;
 
-    return ra_timer_handler(slirp, NULL);
+    ra_timer_handler(slirp, NULL);
 }
 
 void slirp_handle_timer(Slirp *slirp, SlirpTimerId id, void *cb_opaque)
@@ -541,7 +577,8 @@ void slirp_handle_timer(Slirp *slirp, SlirpTimerId id, void *cb_opaque)
 
     switch (id) {
     case SLIRP_TIMER_RA:
-        return ra_timer_handler(slirp, cb_opaque);
+        ra_timer_handler(slirp, cb_opaque);
+        return;
     default:
         abort();
     }
@@ -648,6 +685,14 @@ Slirp *slirp_new(const SlirpConfig *cfg, const SlirpCb *callbacks, void *opaque)
         slirp->cb->init_completed(slirp, slirp->opaque);
     }
 
+    if (cfg->version >= 5) {
+        slirp->mfr_id = cfg->mfr_id;
+        memcpy(slirp->oob_eth_addr, cfg->oob_eth_addr, ETH_ALEN);
+    } else {
+        slirp->mfr_id = 0;
+        memset(slirp->oob_eth_addr, 0, ETH_ALEN);
+    }
+
     ip6_post_init(slirp);
     return slirp;
 }
@@ -701,6 +746,7 @@ void slirp_cleanup(Slirp *slirp)
     ip_cleanup(slirp);
     ip6_cleanup(slirp);
     m_cleanup(slirp);
+    tftp_cleanup(slirp);
 
     g_rand_free(slirp->grand);
 
@@ -739,8 +785,8 @@ static void slirp_update_timeout(Slirp *slirp, uint32_t *timeout)
     *timeout = t;
 }
 
-void slirp_pollfds_fill(Slirp *slirp, uint32_t *timeout,
-                        SlirpAddPollCb add_poll, void *opaque)
+void slirp_pollfds_fill_socket(Slirp *slirp, uint32_t *timeout,
+                               SlirpAddPollSocketCb add_poll, void *opaque)
 {
     struct socket *so, *so_next;
 
@@ -773,7 +819,7 @@ void slirp_pollfds_fill(Slirp *slirp, uint32_t *timeout,
          * NOFDREF can include still connecting to local-host,
          * newly socreated() sockets etc. Don't want to select these.
          */
-        if (so->so_state & SS_NOFDREF || so->s == -1) {
+        if (so->so_state & SS_NOFDREF || not_valid_socket(so->s)) {
             continue;
         }
 
@@ -886,13 +932,79 @@ void slirp_pollfds_fill(Slirp *slirp, uint32_t *timeout,
     slirp_update_timeout(slirp, timeout);
 }
 
+struct PollCbWrap {
+    SlirpAddPollCb add_poll;
+    void *opaque;
+};
+
+static int slirp_pollfds_fill_wrap(slirp_os_socket socket, int events, void *opaque)
+{
+    struct PollCbWrap *wrap = opaque;
+    int fd = (int) socket;
+    if ((slirp_os_socket) fd != socket)
+        g_warning_once("Truncating socket to int failed!");
+    return wrap->add_poll(fd, events, wrap->opaque);
+}
+
+void slirp_pollfds_fill(Slirp *slirp, uint32_t *timeout,
+                        SlirpAddPollCb add_poll, void *opaque)
+{
+    struct PollCbWrap wrap = {
+        .add_poll = add_poll,
+        .opaque = opaque,
+    };
+    slirp_pollfds_fill_socket(slirp, timeout, slirp_pollfds_fill_wrap, &wrap);
+}
+
+void slirp_register_poll_socket(struct socket *so)
+{
+    Slirp *slirp = so->slirp;
+    int fd;
+    if (slirp->cfg_version >= 6 && slirp->cb->register_poll_socket) {
+        slirp->cb->register_poll_socket(so->s, slirp->opaque);
+    } else {
+        fd = (int) so->s;
+        if ((slirp_os_socket) fd != so->s)
+            g_warning_once("Truncating socket to int failed!");
+#ifdef __GNUC__
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
+#endif
+        slirp->cb->register_poll_fd(fd, slirp->opaque);
+#ifdef __GNUC__
+#pragma GCC diagnostic pop
+#endif
+    }
+}
+
+void slirp_unregister_poll_socket(struct socket *so)
+{
+    Slirp *slirp = so->slirp;
+    int fd;
+    if (slirp->cfg_version >= 6 && slirp->cb->unregister_poll_socket) {
+        slirp->cb->unregister_poll_socket(so->s, slirp->opaque);
+    } else {
+        fd = (int) so->s;
+        if ((slirp_os_socket) fd != so->s)
+            g_warning_once("Truncating socket to int failed!");
+#ifdef __GNUC__
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
+#endif
+        slirp->cb->unregister_poll_fd(fd, slirp->opaque);
+#ifdef __GNUC__
+#pragma GCC diagnostic pop
+#endif
+    }
+}
+
 void slirp_pollfds_poll(Slirp *slirp, int select_error,
                         SlirpGetREventsCb get_revents, void *opaque)
 {
     struct socket *so, *so_next;
     int ret;
 
-    curtime = slirp->cb->clock_get_ns(slirp->opaque) / SCALE_MS;
+    curtime = (unsigned int) (slirp->cb->clock_get_ns(slirp->opaque) / SCALE_MS);
 
     /*
      * See if anything has timed out
@@ -926,7 +1038,7 @@ void slirp_pollfds_poll(Slirp *slirp, int select_error,
                 revents = get_revents(so->pollfds_idx, opaque);
             }
 
-            if (so->so_state & SS_NOFDREF || so->s == -1) {
+            if (so->so_state & SS_NOFDREF || not_valid_socket(so->s)) {
                 continue;
             }
 
@@ -1035,7 +1147,7 @@ void slirp_pollfds_poll(Slirp *slirp, int select_error,
                 revents = get_revents(so->pollfds_idx, opaque);
             }
 
-            if (so->s != -1 &&
+            if (have_valid_socket(so->s) &&
                 (revents & (SLIRP_POLL_IN | SLIRP_POLL_HUP | SLIRP_POLL_ERR))) {
                 sorecvfrom(so);
             }
@@ -1054,9 +1166,12 @@ void slirp_pollfds_poll(Slirp *slirp, int select_error,
                 revents = get_revents(so->pollfds_idx, opaque);
             }
 
-            if (so->s != -1 &&
+            if (have_valid_socket(so->s) &&
                 (revents & (SLIRP_POLL_IN | SLIRP_POLL_HUP | SLIRP_POLL_ERR))) {
-                icmp_receive(so);
+                if (so->so_type == IPPROTO_IPV6 || so->so_type == IPPROTO_ICMPV6)
+                    icmp6_receive(so);
+                else
+                    icmp_receive(so);
             }
         }
     }
@@ -1068,9 +1183,9 @@ static void arp_input(Slirp *slirp, const uint8_t *pkt, int pkt_len)
 {
     const struct slirp_arphdr *ah =
         (const struct slirp_arphdr *)(pkt + ETH_HLEN);
-    uint8_t arp_reply[MAX(ETH_HLEN + sizeof(struct slirp_arphdr), 64)];
-    struct ethhdr *reh = (struct ethhdr *)arp_reply;
-    struct slirp_arphdr *rah = (struct slirp_arphdr *)(arp_reply + ETH_HLEN);
+    uint8_t arp_reply[MAX(2 + ETH_HLEN + sizeof(struct slirp_arphdr), 2 + 64)];
+    struct ethhdr *reh = (struct ethhdr *)(arp_reply + 2);
+    struct slirp_arphdr *rah = (struct slirp_arphdr *)(arp_reply + 2 + ETH_HLEN);
     int ar_op;
     struct gfwd_list *ex_ptr;
 
@@ -1123,7 +1238,7 @@ static void arp_input(Slirp *slirp, const uint8_t *pkt, int pkt_len)
             rah->ar_sip = ah->ar_tip;
             memcpy(rah->ar_tha, ah->ar_sha, ETH_ALEN);
             rah->ar_tip = ah->ar_sip;
-            slirp_send_packet_all(slirp, arp_reply, sizeof(arp_reply));
+            slirp_send_packet_all(slirp, arp_reply + 2, sizeof(arp_reply) - 2);
         }
         break;
     case ARPOP_REPLY:
@@ -1152,8 +1267,8 @@ void slirp_input(Slirp *slirp, const uint8_t *pkt, int pkt_len)
         m = m_get(slirp);
         if (!m)
             return;
-        /* Note: we add 2 to align the IP header on 4 bytes,
-         * and add the margin for the tcpiphdr overhead  */
+        /* Note: we add 2 to align the IP header on 8 bytes despite the ethernet
+         * header, and add the margin for the tcpiphdr overhead  */
         if (M_FREEROOM(m) < pkt_len + TCPIPHDR_DELTA + 2) {
             m_inc(m, pkt_len + TCPIPHDR_DELTA + 2);
         }
@@ -1189,9 +1304,9 @@ static int if_encap4(Slirp *slirp, struct mbuf *ifm, struct ethhdr *eh,
     const struct ip *iph = (const struct ip *)ifm->m_data;
 
     if (!arp_table_search(slirp, iph->ip_dst.s_addr, ethaddr)) {
-        uint8_t arp_req[ETH_HLEN + sizeof(struct slirp_arphdr)];
-        struct ethhdr *reh = (struct ethhdr *)arp_req;
-        struct slirp_arphdr *rah = (struct slirp_arphdr *)(arp_req + ETH_HLEN);
+        uint8_t arp_req[2 + ETH_HLEN + sizeof(struct slirp_arphdr)];
+        struct ethhdr *reh = (struct ethhdr *)(arp_req + 2);
+        struct slirp_arphdr *rah = (struct slirp_arphdr *)(arp_req + 2 + ETH_HLEN);
 
         if (!ifm->resolution_requested) {
             /* If the client addr is not known, send an ARP request */
@@ -1218,7 +1333,7 @@ static int if_encap4(Slirp *slirp, struct mbuf *ifm, struct ethhdr *eh,
             /* target IP */
             rah->ar_tip = iph->ip_dst.s_addr;
             slirp->client_ipaddr = iph->ip_dst;
-            slirp_send_packet_all(slirp, arp_req, sizeof(arp_req));
+            slirp_send_packet_all(slirp, arp_req + 2, sizeof(arp_req) - 2);
             ifm->resolution_requested = true;
 
             /* Expire request and drop outgoing packet after 1 second */
@@ -1268,13 +1383,13 @@ static int if_encap6(Slirp *slirp, struct mbuf *ifm, struct ethhdr *eh,
 int if_encap(Slirp *slirp, struct mbuf *ifm)
 {
     uint8_t buf[IF_MTU_MAX + 100];
-    struct ethhdr *eh = (struct ethhdr *)buf;
+    struct ethhdr *eh = (struct ethhdr *)(buf + 2);
     uint8_t ethaddr[ETH_ALEN];
     const struct ip *iph = (const struct ip *)ifm->m_data;
     int ret;
     char ethaddr_str[ETH_ADDRSTRLEN];
 
-    if (ifm->m_len + ETH_HLEN > sizeof(buf)) {
+    if (ifm->m_len + ETH_HLEN > sizeof(buf) - 2) {
         return 1;
     }
 
@@ -1302,8 +1417,8 @@ int if_encap(Slirp *slirp, struct mbuf *ifm)
                                            sizeof(ethaddr_str)));
     DEBUG_ARG("dst = %s", slirp_ether_ntoa(eh->h_dest, ethaddr_str,
                                            sizeof(ethaddr_str)));
-    memcpy(buf + sizeof(struct ethhdr), ifm->m_data, ifm->m_len);
-    slirp_send_packet_all(slirp, buf, ifm->m_len + ETH_HLEN);
+    memcpy(buf + 2 + sizeof(struct ethhdr), ifm->m_data, ifm->m_len);
+    slirp_send_packet_all(slirp, buf + 2, ifm->m_len + ETH_HLEN);
     return 1;
 }
 
@@ -1325,7 +1440,7 @@ int slirp_remove_hostfwd(Slirp *slirp, int is_udp, struct in_addr host_addr,
             addr.sin_family == AF_INET &&
             addr.sin_addr.s_addr == host_addr.s_addr &&
             addr.sin_port == port) {
-            so->slirp->cb->unregister_poll_fd(so->s, so->slirp->opaque);
+            slirp_unregister_poll_socket(so);
             closesocket(so->s);
             sofree(so);
             return 0;
@@ -1367,7 +1482,7 @@ int slirp_remove_hostxfwd(Slirp *slirp,
         if ((so->so_state & SS_HOSTFWD) &&
             getsockname(so->s, (struct sockaddr *)&addr, &addr_len) == 0 &&
             sockaddr_equal(&addr, (const struct sockaddr_storage *) haddr)) {
-            so->slirp->cb->unregister_poll_fd(so->s, so->slirp->opaque);
+            slirp_unregister_poll_socket(so);
             closesocket(so->s);
             sofree(so);
             return 0;
@@ -1503,16 +1618,16 @@ int slirp_remove_guestfwd(Slirp *slirp, struct in_addr guest_addr,
                            htons(guest_port));
 }
 
-ssize_t slirp_send(struct socket *so, const void *buf, size_t len, int flags)
+slirp_ssize_t slirp_send(struct socket *so, const void *buf, size_t len, int flags)
 {
-    if (so->s == -1 && so->guestfwd) {
+    if (not_valid_socket(so->s) && so->guestfwd) {
         /* XXX this blocks entire thread. Rewrite to use
          * qemu_chr_fe_write and background I/O callbacks */
         so->guestfwd->write_cb(buf, len, so->guestfwd->opaque);
         return len;
     }
 
-    if (so->s == -1) {
+    if (not_valid_socket(so->s)) {
         /*
          * This should in theory not happen but it is hard to be
          * sure because some code paths will end up with so->s == -1
@@ -1579,7 +1694,17 @@ void slirp_socket_recv(Slirp *slirp, struct in_addr guest_addr, int guest_port,
 
 void slirp_send_packet_all(Slirp *slirp, const void *buf, size_t len)
 {
-    ssize_t ret = slirp->cb->send_packet(buf, len, slirp->opaque);
+    slirp_ssize_t ret;
+
+    if (len < ETH_MINLEN) {
+        char tmp[ETH_MINLEN];
+        memcpy(tmp, buf, len);
+        memset(tmp + len, 0, ETH_MINLEN - len);
+
+        ret = slirp->cb->send_packet(tmp, ETH_MINLEN, slirp->opaque);
+    } else {
+        ret = slirp->cb->send_packet(buf, len, slirp->opaque);
+    }
 
     if (ret < 0) {
         g_critical("Failed to send packet, ret: %ld", (long)ret);
