@@ -16,8 +16,8 @@ static void sofcantrcvmore(struct socket *so);
 static void sofcantsendmore(struct socket *so);
 
 struct socket *solookup(struct socket **last, struct socket *head,
-                        struct sockaddr_storage *lhost,
-                        struct sockaddr_storage *fhost)
+                        const struct sockaddr_storage *lhost,
+                        const struct sockaddr_storage *fhost)
 {
     struct socket *so = *last;
 
@@ -50,8 +50,8 @@ struct socket *socreate(Slirp *slirp, int type)
     memset(so, 0, sizeof(struct socket));
     so->so_type = type;
     so->so_state = SS_NOFDREF;
-    so->s = -1;
-    so->s_aux = -1;
+    so->s = SLIRP_INVALID_SOCKET;
+    so->s_aux = SLIRP_INVALID_SOCKET;
     so->slirp = slirp;
     so->pollfds_idx = -1;
 
@@ -61,17 +61,17 @@ struct socket *socreate(Slirp *slirp, int type)
 /*
  * Remove references to so from the given message queue.
  */
-static void soqfree(struct socket *so, struct slirp_quehead *qh)
+static void soqfree(const struct socket *so, struct slirp_quehead *qh)
 {
     struct mbuf *ifq;
 
     for (ifq = (struct mbuf *)qh->qh_link; (struct slirp_quehead *)ifq != qh;
-         ifq = ifq->ifq_next) {
-        if (ifq->ifq_so == so) {
+         ifq = ifq->m_next) {
+        if (ifq->m_so == so) {
             struct mbuf *ifm;
-            ifq->ifq_so = NULL;
-            for (ifm = ifq->ifs_next; ifm != ifq; ifm = ifm->ifs_next) {
-                ifm->ifq_so = NULL;
+            ifq->m_so = NULL;
+            for (ifm = ifq->m_nextpkt; ifm != ifq; ifm = ifm->m_nextpkt) {
+                ifm->m_so = NULL;
             }
         }
     }
@@ -84,7 +84,7 @@ void sofree(struct socket *so)
 {
     Slirp *slirp = so->slirp;
 
-    if (so->s_aux != -1) {
+    if (have_valid_socket(so->s_aux)) {
         closesocket(so->s_aux);
     }
 
@@ -214,7 +214,7 @@ int soread(struct socket *so)
             }
 
             DEBUG_MISC(" --- soread() disconnected, nn = %d, errno = %d-%s", nn,
-                       errno, strerror(errno));
+                       errno, g_strerror(errno));
             sofcantrcvmore(so);
 
             if (err == ECONNABORTED || err == ECONNRESET || err == ECONNREFUSED ||
@@ -577,14 +577,36 @@ void sorecvfrom(struct socket *so)
             else if (errno == ENETUNREACH)
                 code = ICMP_UNREACH_NET;
 
-            DEBUG_MISC(" udp icmp rx errno = %d-%s", errno, strerror(errno));
-            icmp_send_error(so->so_m, ICMP_UNREACH, code, 0, strerror(errno));
+            DEBUG_MISC(" udp icmp rx errno = %d-%s", errno, g_strerror(errno));
+            icmp_send_error(so->so_m, ICMP_UNREACH, code, 0, g_strerror(errno));
         } else {
             icmp_reflect(so->so_m);
             so->so_m = NULL; /* Don't m_free() it again! */
         }
         /* No need for this socket anymore, udp_detach it */
         udp_detach(so);
+    } else if (so->so_type == IPPROTO_ICMPV6) { /* This is a "ping" reply */
+        int len;
+
+        len = recvfrom(so->s, buff, 256, 0, (struct sockaddr *)&addr, &addrlen);
+        /* XXX Check if reply is "correct"? */
+
+        if (len == -1 || len == 0) {
+            uint8_t code = ICMP6_UNREACH_PORT;
+
+            if (errno == EHOSTUNREACH)
+                code = ICMP6_UNREACH_ADDRESS;
+            else if (errno == ENETUNREACH)
+                code = ICMP6_UNREACH_NO_ROUTE;
+
+            DEBUG_MISC(" udp icmp6 rx errno = %d-%s", errno, g_strerror(errno));
+            icmp6_send_error(so->so_m, ICMP_UNREACH, code);
+        } else {
+            icmp6_reflect(so->so_m);
+            so->so_m = NULL; /* Don't m_free() it again! */
+        }
+        /* No need for this socket anymore, udp_detach it */
+        udp_detach(so);
     } else { /* A "normal" UDP packet */
         struct mbuf *m;
         int len;
@@ -595,7 +617,7 @@ void sorecvfrom(struct socket *so)
 #endif
 
         if (ioctlsocket(so->s, FIONREAD, &n) != 0) {
-            DEBUG_MISC(" ioctlsocket errno = %d-%s\n", errno, strerror(errno));
+            DEBUG_MISC(" ioctlsocket errno = %d-%s\n", errno, g_strerror(errno));
             return;
         }
 
@@ -632,40 +654,48 @@ void sorecvfrom(struct socket *so)
         m->m_len = recvfrom(so->s, m->m_data, len, 0, (struct sockaddr *)&addr,
                             &addrlen);
         DEBUG_MISC(" did recvfrom %d, errno = %d-%s", m->m_len, errno,
-                   strerror(errno));
-        if (m->m_len < 0) {
-            /* Report error as ICMP */
-            switch (so->so_lfamily) {
-                uint8_t code;
-            case AF_INET:
-                code = ICMP_UNREACH_PORT;
+                   g_strerror(errno));
+        if (m->m_len < 0) {    	
+            if (errno == ENOTCONN) {
+                /*
+                 * UDP socket got burnt, e.g. by suspend on iOS. Tear it down
+                 * and let it get re-created if the guest still needs it
+                 */
+                udp_detach(so);
+            } else {
+                /* Report error as ICMP */
+                switch (so->so_lfamily) {
+                    uint8_t code;
+                case AF_INET:
+                    code = ICMP_UNREACH_PORT;
 
-                if (errno == EHOSTUNREACH) {
-                    code = ICMP_UNREACH_HOST;
-                } else if (errno == ENETUNREACH) {
-                    code = ICMP_UNREACH_NET;
+                    if (errno == EHOSTUNREACH) {
+                        code = ICMP_UNREACH_HOST;
+                    } else if (errno == ENETUNREACH) {
+                        code = ICMP_UNREACH_NET;
+                    }
+
+                    DEBUG_MISC(" rx error, tx icmp ICMP_UNREACH:%i", code);
+                    icmp_send_error(so->so_m, ICMP_UNREACH, code, 0,
+                                    g_strerror(errno));
+                    break;
+                case AF_INET6:
+                    code = ICMP6_UNREACH_PORT;
+
+                    if (errno == EHOSTUNREACH) {
+                        code = ICMP6_UNREACH_ADDRESS;
+                    } else if (errno == ENETUNREACH) {
+                        code = ICMP6_UNREACH_NO_ROUTE;
+                    }
+
+                    DEBUG_MISC(" rx error, tx icmp6 ICMP_UNREACH:%i", code);
+                    icmp6_send_error(so->so_m, ICMP6_UNREACH, code);
+                    break;
+                default:
+                    g_assert_not_reached();
                 }
-
-                DEBUG_MISC(" rx error, tx icmp ICMP_UNREACH:%i", code);
-                icmp_send_error(so->so_m, ICMP_UNREACH, code, 0,
-                                strerror(errno));
-                break;
-            case AF_INET6:
-                code = ICMP6_UNREACH_PORT;
-
-                if (errno == EHOSTUNREACH) {
-                    code = ICMP6_UNREACH_ADDRESS;
-                } else if (errno == ENETUNREACH) {
-                    code = ICMP6_UNREACH_NO_ROUTE;
-                }
-
-                DEBUG_MISC(" rx error, tx icmp6 ICMP_UNREACH:%i", code);
-                icmp6_send_error(so->so_m, ICMP6_UNREACH, code);
-                break;
-            default:
-                g_assert_not_reached();
+                m_free(m);
             }
-            m_free(m);
         } else {
             /*
              * Hack: domain name lookup will be used the most for UDP,
@@ -761,17 +791,14 @@ int sosendto(struct socket *so, struct mbuf *m)
     return 0;
 }
 
-/*
- * Listen for incoming TCP connections
- * On failure errno contains the reason.
- */
 struct socket *tcpx_listen(Slirp *slirp,
                            const struct sockaddr *haddr, socklen_t haddrlen,
                            const struct sockaddr *laddr, socklen_t laddrlen,
                            int flags)
 {
     struct socket *so;
-    int s, opt = 1;
+    slirp_os_socket s;
+    int opt = 1;
     socklen_t addrlen;
 
     DEBUG_CALL("tcpx_listen");
@@ -837,22 +864,19 @@ struct socket *tcpx_listen(Slirp *slirp,
     sockaddr_copy(&so->lhost.sa, sizeof(so->lhost), laddr, laddrlen);
 
     s = slirp_socket(haddr->sa_family, SOCK_STREAM, 0);
-    if ((s < 0) ||
+    if ((not_valid_socket(s)) ||
         (haddr->sa_family == AF_INET6 && slirp_socket_set_v6only(s, (flags & SS_HOSTFWD_V6ONLY) != 0) < 0) ||
         (slirp_socket_set_fast_reuse(s) < 0) ||
         (bind(s, haddr, haddrlen) < 0) ||
         (listen(s, 1) < 0)) {
         int tmperrno = errno; /* Don't clobber the real reason we failed */
-        if (s >= 0) {
+        if (have_valid_socket(s)) {
             closesocket(s);
         }
         sofree(so);
         /* Restore the real errno */
-#ifdef _WIN32
-        WSASetLastError(tmperrno);
-#else
         errno = tmperrno;
-#endif
+
         return NULL;
     }
     setsockopt(s, SOL_SOCKET, SO_OOBINLINE, &opt, sizeof(int));
@@ -863,6 +887,8 @@ struct socket *tcpx_listen(Slirp *slirp,
     sotranslate_accept(so);
 
     so->s = s;
+    slirp_register_poll_socket(so);
+
     return so;
 }
 
@@ -931,10 +957,6 @@ static void sofcantsendmore(struct socket *so)
     }
 }
 
-/*
- * Set write drain mode
- * Set CANTSENDMORE once all data has been write()n
- */
 void sofwdrain(struct socket *so)
 {
     if (so->so_rcv.sb_cc)
@@ -985,9 +1007,6 @@ static bool sotranslate_out6(Slirp *s, struct socket *so, struct sockaddr_in6 *s
 }
 
 
-/*
- * Translate addr in host addr when it is a virtual address
- */
 int sotranslate_out(struct socket *so, struct sockaddr_storage *addr)
 {
     bool ok = true;
@@ -1045,9 +1064,6 @@ void sotranslate_in(struct socket *so, struct sockaddr_storage *addr)
     }
 }
 
-/*
- * Translate connections from localhost to the real hostname
- */
 void sotranslate_accept(struct socket *so)
 {
     Slirp *slirp = so->slirp;
@@ -1073,7 +1089,7 @@ void sotranslate_accept(struct socket *so)
          * this source port by binding to port 0 so that the OS allocates a
          * port for us. If this fails, we fall back to choosing a random port
          * with a random number generator. */
-        int s;
+        slirp_os_socket s;
         struct sockaddr_in in_addr;
         struct sockaddr_in6 in6_addr;
         socklen_t in_addr_len;
@@ -1094,7 +1110,7 @@ void sotranslate_accept(struct socket *so)
                 g_assert_not_reached();
                 break;
             }
-            if (s < 0) {
+            if (not_valid_socket(s)) {
                 g_error("Ephemeral slirp_socket() allocation failed");
                 goto unix2inet_cont;
             }
@@ -1137,7 +1153,7 @@ unix2inet_cont:
                 g_assert_not_reached();
                 break;
             }
-            if (s < 0) {
+            if (not_valid_socket(s)) {
                 g_error("Ephemeral slirp_socket() allocation failed");
                 goto unix2inet6_cont;
             }
