@@ -29,6 +29,11 @@
 #include <sys/stat.h>
 #include <fcntl.h>
 
+#if defined(_WIN32)
+#include <io.h>
+#include <share.h>
+#endif
+
 static inline int tftp_session_in_use(struct tftp_session *spt)
 {
     return (spt->slirp != NULL);
@@ -297,13 +302,13 @@ static void tftp_handle_rrq(Slirp *slirp, struct sockaddr_storage *srcsas,
 
     /* check if a session already exists and if so terminate it */
     s = tftp_session_find(slirp, srcsas, &tp->hdr);
-    if (s >= 0) {
+    if (have_valid_socket(s)) {
         tftp_session_terminate(&slirp->tftp_sessions[s]);
     }
 
     s = tftp_session_allocate(slirp, srcsas, &tp->hdr);
 
-    if (s < 0) {
+    if (not_valid_socket(s)) {
         return;
     }
 
@@ -437,7 +442,7 @@ static void tftp_handle_ack(Slirp *slirp, struct sockaddr_storage *srcsas,
 
     s = tftp_session_find(slirp, srcsas, hdr);
 
-    if (s < 0) {
+    if (not_valid_socket(s)) {
         return;
     }
 
@@ -451,7 +456,7 @@ static void tftp_handle_error(Slirp *slirp, struct sockaddr_storage *srcsas,
 
     s = tftp_session_find(slirp, srcsas, hdr);
 
-    if (s < 0) {
+    if (not_valid_socket(s)) {
         return;
     }
 
